"""This subpackage provides a Python interface for SRIM for simulating collisional cascades."""

from irradiapy.srim import analysis
from irradiapy.srim.srimdb import SRIMDB
from irradiapy.srim.target import Element, Layer, Target
